@@ -9,7 +9,6 @@
 """
 from types import FunctionType as Function
 from typing import Any, Callable, Dict, List, Type, TypeVar, Union, cast
-import warnings
 
 DataClass = Any  # type hint for variables that should be data class instances
 
@@ -70,13 +69,15 @@ class DataClassMeta(type):
         dict_.pop('__dict__', None)
         dict_ = {f: v for f, v in dict_.items() if type(v).__name__ != 'member_descriptor'}
 
-        # collect annotations, defaults, slots and options from this class' ancestors, in definition order
+        # collect functions, annotations, defaults, slots and options from this class' ancestors, in definition order
 
         all_annotations = {}
         all_defaults = {}
         all_slots = set()
         options = dict(mcs.DEFAULT_OPTIONS)
-        post_init = False
+
+        # record all functions defined by the user up through the inheritance chain
+        all_attrs = {a for b in bases for a in dir(b) if is_user_func(getattr(b, a))} | dict_.keys()
 
         dataclass_bases = [vars(b) for b in bases if hasattr(b, '__dataclass__')]
         for b in dataclass_bases + [dict_]:
@@ -85,11 +86,7 @@ class DataClassMeta(type):
             all_slots.update(b.get('__slots__', set()))
             options.update(b.get('__dataclass__', {}))
 
-            # add user-defined methods to dict_ so that we know not to replace them. Do not replace methods
-            # that are already there as this would break overriding
-            dict_.update({f: v for f, v in b.items() if f not in dict_ and is_user_func(v)})
-
-            post_init = is_user_func(b.get('__init__')) or is_user_func(b.get('__post_init__'))
+        post_init = '__post_init__' in all_attrs
 
         # update options and defaults for *this* class
 
@@ -114,31 +111,31 @@ class DataClassMeta(type):
             # if the slots option gets removed, remove __slots__
             del dict_['__slots__']
 
-        if options['init'] and all_annotations:  # only generate __init__ if there are fields to set
-            if post_init and is_user_func(dict_.get('__init__')):
-                warnings.warn('Defining post-initialisation logic in __init__ is deprecated and will be removed in a'
-                              ' forthcoming version. Please rename your method to __post_init__.', DeprecationWarning)
-                dict_['__post_init__'] = dict_.pop('__init__')
-            dict_['__init__'] = generate_init(all_annotations, all_defaults, post_init,
-                                              options['kwargs'], options['frozen'])
+        if options['init'] and all_annotations and '__init__' not in all_attrs:
+            dict_.setdefault('__init__', generate_init(all_annotations, all_defaults, post_init,
+                                                       options['kwargs'], options['frozen']))
+        elif options['init'] and '__init__' not in dict_ and '__post_init__' in dict_:
+            # if no fields to set, alias __post_init__ to __post_init__ so it is always called (unless init=False)
+            dict_['__init__'] = dict_['__post_init__']
 
         if options['repr']:
-            dict_.setdefault('__repr__', __repr__)
+            '__repr__' in all_attrs or dict_.setdefault('__repr__', __repr__)
 
         if options['eq']:
-            dict_.setdefault('__eq__', __eq__)
+            '__eq__' in all_attrs or dict_.setdefault('__eq__', __eq__)
 
         if options['iter']:
-            dict_.setdefault('__iter__', __iter__)
+            '__iter__' in all_attrs or dict_.setdefault('__iter__', __iter__)
 
         if options['frozen']:
-            dict_['__delattr__'] = dict_['__setattr__'] = __setattr__
+            '__delattr__' in all_attrs or dict_.setdefault('__delattr__', __setattr__)
+            '__setattr__' in all_attrs or dict_.setdefault('__setattr__', __setattr__)
 
         if options['order']:
-            dict_.setdefault('__lt__', __lt__)
+            '__lt__' in all_attrs or dict_.setdefault('__lt__', __lt__)
 
         if (options['eq'] and options['frozen']) or options['unsafe_hash']:
-            dict_.setdefault('__hash__', generate_hash(all_annotations))
+            '__hash__' in all_attrs or dict_.setdefault('__hash__', generate_hash(all_annotations))
 
         return super().__new__(mcs, name, bases, dict_)
 
@@ -202,10 +199,11 @@ def eval_function(name: str, lines: List[str], annotations: Dict, locals_: Dict,
     return function
 
 
-def is_user_func(obj: Any) -> bool:
-    """Given an object, returns it if it is an object explicitly defined by the user (i.e. not generated by
-    dataclassy and not a wrapper/descriptor). Otherwise, returns None."""
-    return type(obj) is Function and not hasattr(obj, '__dataclass__')
+def is_user_func(obj: Any, object_methods=frozenset(vars(object).values())) -> bool:
+    """Given an object, returns true if it is a function explicitly defined by the user (i.e. not generated by
+    dataclassy and not a wrapper/descriptor). Otherwise, returns False.
+    The unusual check against object_methods is because PyPy returns True for methods on object (unlike CPython)."""
+    return type(obj) is Function and obj not in object_methods and not hasattr(obj, '__dataclass__')
 
 
 # generic method implementations common to all data classes
